package account

import (
	"context"
	"crypto/tls"
	"crypto/x509"
	"fmt"
	"net/url"
	"time"

	"git.perx.ru/perxis/perxis-go/pkg/cache"
	serviceMembers "git.perx.ru/perxis/perxis-go/pkg/members/middleware"
	membersObserverTransport "git.perx.ru/perxis/perxis-go/pkg/members/observer/transport/grpc"
	membersTransport "git.perx.ru/perxis/perxis-go/pkg/members/transport/grpc"
	serviceOrganizations "git.perx.ru/perxis/perxis-go/pkg/organizations/middleware"
	organizationsTransport "git.perx.ru/perxis/perxis-go/pkg/organizations/transport/grpc"
	serviceUsers "git.perx.ru/perxis/perxis-go/pkg/users/middleware"
	usersTransport "git.perx.ru/perxis/perxis-go/pkg/users/transport/grpc"
	"go.opentelemetry.io/contrib/instrumentation/google.golang.org/grpc/otelgrpc"
	"go.uber.org/zap"
	"golang.org/x/oauth2/clientcredentials"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials"
	"google.golang.org/grpc/credentials/insecure"
	"google.golang.org/grpc/credentials/oauth"
)

const (
	DefaultCacheSize = 1000
	DefaultCacheTTL  = time.Second * 10
)

func NewClient(ctx context.Context, addr string, opts ...Option) (*Account, *grpc.ClientConn, error) {

	client := &Account{}
	dialOpts := make([]grpc.DialOption, 0)
	config := &config{}

	for _, o := range opts {
		o(config)

	}

	if config.logger == nil {
		config.logger = zap.NewNop()
	}

	dialOpts = append(dialOpts, grpc.WithUnaryInterceptor(otelgrpc.UnaryClientInterceptor()),
		grpc.WithStreamInterceptor(otelgrpc.StreamClientInterceptor()))

	if config.auth == nil {
		dialOpts = append(dialOpts, grpc.WithTransportCredentials(insecure.NewCredentials()))
	} else {
		if config.auth.TLS != nil {

			certPool := x509.NewCertPool()

			if !certPool.AppendCertsFromPEM(config.auth.TLS.cacert) {
				return nil, nil, fmt.Errorf("CA certificate not loaded")
			}

			clientCert, err := tls.X509KeyPair(config.auth.TLS.cert, config.auth.TLS.key)
			if err != nil {
				return nil, nil, err

			}

			tlsConfig := &tls.Config{
				Certificates: []tls.Certificate{clientCert},
				RootCAs:      certPool,
			}

			dialOpts = append(dialOpts,
				grpc.WithTransportCredentials(credentials.NewTLS(tlsConfig)),
			)
		}

		if config.auth.OAuth2 != nil {
			if config.auth.OAuth2 != nil {
				// create external grpc client
				conf := &clientcredentials.Config{
					TokenURL:       config.auth.OAuth2.tokenURL,
					ClientID:       config.auth.OAuth2.clientID,
					ClientSecret:   config.auth.OAuth2.clientSecret,
					EndpointParams: url.Values{"audience": {config.auth.OAuth2.audience}},
				}
				cred := oauth.TokenSource{
					TokenSource: conf.TokenSource(ctx),
				}
				dialOpts = append(dialOpts,
					grpc.WithPerRPCCredentials(cred),
					grpc.WithTransportCredentials(credentials.NewTLS(&tls.Config{InsecureSkipVerify: true})), // обязательно использовать tls для credentials oauth.TokenSource https://github.com/grpc/grpc-go/blob/64031cbfcf4d84c026be93ad7b74b3c290100893/credentials/oauth/oauth.go#L160
				)
			}
		}

	}

	accountConn, err := grpc.Dial(addr, dialOpts...)
	if err != nil {
		return nil, nil, err
	}

	client.Members = membersTransport.NewGRPCClient(accountConn, "", config.clientOptions...)
	client.Organizations = organizationsTransport.NewGRPCClient(accountConn, "", config.clientOptions...)
	client.Users = usersTransport.NewGRPCClient(accountConn, "", config.clientOptions...)
	client.MembersObserver = membersObserverTransport.NewGRPCClient(accountConn, "", config.clientOptions...)

	if !config.noCache {
		client = WithCaching(client, DefaultCacheSize, DefaultCacheTTL)
	}

	if !config.noLog {
		client = WithLogging(client, config.logger, config.accessLog)
	}

	return client, accountConn, nil
}

func WithCaching(client *Account, size int, ttl time.Duration) *Account {
	c := *client

	c.Members = serviceMembers.CachingMiddleware(cache.NewCache(size, ttl))(client.Members)
	c.Organizations = serviceOrganizations.CachingMiddleware(cache.NewCache(size, ttl))(client.Organizations)
	c.Users = serviceUsers.CachingMiddleware(cache.NewCache(size, ttl))(client.Users)

	return &c
}

func WithLogging(client *Account, logger *zap.Logger, accessLog bool) *Account {
	c := *client

	c.Members = serviceMembers.WithLog(c.Members, logger, accessLog)
	c.Organizations = serviceOrganizations.WithLog(c.Organizations, logger, accessLog)
	c.Users = serviceUsers.WithLog(c.Users, logger, accessLog)

	return &c
}
