package content

import (
	kitgrpc "github.com/go-kit/kit/transport/grpc"
	"go.uber.org/zap"
)

type Config struct {
	Auth      *AuthConfig
	NoCache   bool
	NoLog     bool
	AccessLog bool
	Debug     bool
	NoDecode  bool

	ClientOptions []kitgrpc.ClientOption

	Logger *zap.Logger
}

type AuthConfig struct {
	*OAuth2  `json:"oauth_2,omitempty"`
	*APIKey  `json:"api_key,omitempty"`
	*TLS     `json:"tls,omitempty"`
	Insecure bool `json:"insecure,omitempty"`
}

type OAuth2 struct {
	TokenURL     string `json:"token_url,omitempty"`
	ClientID     string `json:"client_id,omitempty"` // параметр из auth0 (клиент с таким id должен быть создан в perxis)
	ClientSecret string `json:"client_secret,omitempty"`
	Audience     string `json:"audience,omitempty"` // параметр из auth0 (название связанного с Application API)
}

type TLS struct {
	CaCert     []byte `json:"tls-cacert"`
	Cert       []byte `json:"tls-cert"`
	Key        []byte `json:"tls-key"`
	SkipVerify bool
}

type APIKey struct {
	APIKey string `json:"api_key"`
}

type Option func(c *Config)

func AuthOAuth2(tokenUrl, clientID, clientSecret, audience string) Option {
	return func(c *Config) {
		if c.Auth == nil {
			c.Auth = &AuthConfig{}
		}
		c.Auth.OAuth2 = &OAuth2{
			TokenURL:     tokenUrl,
			ClientID:     clientID,
			ClientSecret: clientSecret,
			Audience:     audience,
		}
	}
}

func AuthTLS(cacert, cert, key []byte) Option {
	return func(c *Config) {
		if c.Auth == nil {
			c.Auth = &AuthConfig{}
		}
		c.Auth.TLS = &TLS{
			CaCert: cacert,
			Cert:   cert,
			Key:    key,
		}
	}
}

func AuthAPIKey(key string) Option {
	return func(c *Config) {
		if c.Auth == nil {
			c.Auth = &AuthConfig{}
		}
		c.Auth.APIKey = &APIKey{APIKey: key}
	}
}

func AuthInsecure() Option {
	return func(c *Config) {
		if c.Auth == nil {
			c.Auth = &AuthConfig{}
		}
		c.Auth.Insecure = true
	}
}

func Auth(cfg *AuthConfig) Option {
	return func(c *Config) {
		c.Auth = cfg
	}
}

func NoCache() Option {
	return func(c *Config) {
		c.NoCache = true
	}
}

func NoLog() Option {
	return func(c *Config) {
		c.NoLog = true
	}
}

func Logger(logger *zap.Logger) Option {
	return func(c *Config) {
		c.Logger = logger
	}
}

func AccessLog() Option {
	return func(c *Config) {
		c.AccessLog = true
	}
}

func GrpcClientOptions(opts ...kitgrpc.ClientOption) Option {
	return func(c *Config) {
		c.ClientOptions = opts
	}
}
